//[[
#include <ivl/ivl>
#include <ivl/cv>
namespace forests {
using namespace ivl;  using ivl::rand;
typedef image<double>::const_box patch;
const int alpha = 9;              // patch side
const int sz = 3 * _[alpha]->*2;  // patch size
struct sample { patch p;  int k; };
typedef array<sample>::const_indirect_array data;
struct test {
	int type, feat[2];  double th;
	double f(const sample& x) const {
		double v1 = x.p[feat[0]], v2 = x.p[feat[1]];
		return type == 0 ? v1 : type == 1 ? v1 + v2 :
		       type == 2 ? v1 - v2 : abs(v1 - v2);
	}
	void gen(const data& X) {
		type = rand(0, 3);  lnk(feat) = rand(2, 0, sz-1);
		th = rand[extrema( _[this][&test::f](X) )];
	}
};
const int K = 24;  // number of classes
typedef array<double, fixed<K> > prob;
struct node
	{ virtual const prob& classify(const sample&) = 0; };
typedef node* tree;
struct split_node : public node {
	test t;  node *l, *r;
	split_node(test t,node *l,node *r): t(t),l(l),r(r) {}
	const prob& classify(const sample& x)
		{ return (t.f(x) < t.th ? l : r)->classify(x); }
};
struct leaf_node : public node {
	prob P;  leaf_node(const prob& P) : P(P) {}
	const prob& classify(const sample&) { return P; }
};
typedef array<tree> forest;
const int T = 4;  // number of trees
int classify_(const forest& F, const sample& x)
	{ return arg_max(sum( F[&node::classify](x) )); }
static binary_elem<int, const forest&, const sample&, classify_> classify;
ret<data, data> split_(const test& t, const data& X) {
	array<bool> mask = _[t][&test::f](X) < t.th;
	return (_, X[find(mask)], X[find(!mask)]);
}
static binary_elem<ret<data, data>, const test&, const data&, split_> split;
prob hist(const array<int>& k)
	{ prob h(K, 0.0);  ++h[k];  return h / k.size(); }
double entropy(const prob& P)
	{ return -sum( (P * log(P))[P>0] ); }
double H(const data& X)
	{ return entropy(hist( X[&sample::k] )); }
double gain_(const data& L, const data& R) {
	int l = L.size(), r = R.size();
	return !(l&&r) ? -infty : -(l*H(L) + r*H(R)) / (l+r);
}
const double gamma = -0.1;  // gain threshold
static binary_elem<double, const data&, const data&, gain_> gain;
const int S = 10;  // number of random tests / splits
node* learn(const data& X) {
	array<test> tests(S);  tests[&test::gen](_[X]);
	double g;  int best;  // max gain, best test
	(_, g, best) = max++( gain[split(tests, _[X])] );
	if(g < gamma && !isinf(g)) {
		data L, R;  const test& t = tests[best];
		(_, L, R) = split(t, _[X]);
		return new split_node(t, learn(L), learn(R));
	}
	return new leaf_node(hist( X[&sample::k] ));
}
static unary_elem<tree, const data&, learn, loops::intensive> core_learn;
forest learn(const array<sample>& X)
	{ return core_learn(rep(T, data(X))); }
struct group { image<double> im;  array_2d<int> id; };
sample pick(const point<int>& q, const group& g) {
	sample x;  int h = alpha / 2;  // half-side
	x.p = g.im((q.r-h,_,q.r+h), (q.c-h,_,q.c+h), _);
	if(!g.id.empty()) x.k = g.id(q);
	return x;
}
const int sigma = 5;  // grid step
array_2d<sample> sparse(const group& g, int s = sigma) {
	int h = alpha / 2;  // half-side
	size_range rows = (h,_,s,_,g.im.rows()-h-1);
	size_range cols = (h,_,s,_,g.im.cols()-h-1);
	return _[pick]( pnt[grid(rows, cols)], _[g] );
}
array_2d<sample> dense(const group& g)
	{ return sparse(g, 1); }
const int c[K] =
	{  0,  2,  8, 10, 32, 34, 40, 42,  1,  3,  9, 11,
	  33, 35, 41, 43,  4,  6, 12, 38, 44, 46,  5,  7  };
const array<int> colors(c);
array<int> labels() {
	array<int> id(max(colors) + 1, 0);
	id[colors] = (0,_,K-1);  return id;
}
array_2d<int> im2label(const image<unsigned char>& im) {
	return labels() [ im(_,_,0) /  4 +
	                  im(_,_,1) / 16 + im(_,_,2) / 64 ];
}
image<unsigned char> label2im(const array_2d<int>& id) {
	image<unsigned char> im(id.rows(), id.cols(), 3);
	array_2d<int> c = colors[id];
	im(_,_,0) = c >> 4;  im(_,_,2) = c;
	im(_,_,1) = c >> 2;  im &= 3;  return im <<= 6;
}
}  // namespace forests
int main(int argc, char** argv) {
	using namespace ivl;
	using namespace forests;
	group g;  g.im = imread<double>(argv[1]);
	g.id = im2label(imread<unsigned char>(argv[2]));
	imshow(g.im, "training image");
	imshow(label2im(g.id), "ground truth");
	forest F = learn(sparse(g));
	g.im.load(argv[3]);  g.id = _;
	imshow(g.im, "test image");
	g.id = classify(F, dense(g));
	imshow(label2im(g.id), "classification");
}
//>
